From ddfba877e3c386598e3790b674136717fc0b0321 Mon Sep 17 00:00:00 2001
From: Craig Ringer <craig.ringer@2ndquadrant.com>
Date: Tue, 29 Sep 2020 11:43:02 +0800
Subject: [PATCH 1/2] Add libpq_be function to report bytes pending send in
 internal buffer

We have pq_is_send_pending() to report whether there's anything waiting
in the libpq send buffer, but not how much.

For stats reporting, add pq_nbytes_pending() so we can see how much is waiting
to send/receive. The send buffer is the main one of interest, as in general the
receive buffer is small and transient, but it can still matter for some
workloads like logical receivers.
---
 src/backend/libpq/pqcomm.c | 35 ++++++++++++++++++++++++++++++++++-
 src/backend/libpq/pqmq.c   |  3 ++-
 src/include/libpq/libpq.h  | 27 +++++++++++++++++++++++++++
 3 files changed, 63 insertions(+), 2 deletions(-)

diff --git a/src/backend/libpq/pqcomm.c b/src/backend/libpq/pqcomm.c
index ac986c0505..f7daaebe6c 100644
--- a/src/backend/libpq/pqcomm.c
+++ b/src/backend/libpq/pqcomm.c
@@ -160,6 +160,7 @@ static int	socket_putmessage(char msgtype, const char *s, size_t len);
 static void socket_putmessage_noblock(char msgtype, const char *s, size_t len);
 static void socket_startcopyout(void);
 static void socket_endcopyout(bool errorAbort);
+static int socket_nbytes_pending(size_t * rxsz, size_t * txsz);
 static int	internal_putbytes(const char *s, size_t len);
 static int	internal_flush(void);
 
@@ -176,7 +177,8 @@ static const PQcommMethods PqCommSocketMethods = {
 	socket_putmessage,
 	socket_putmessage_noblock,
 	socket_startcopyout,
-	socket_endcopyout
+	socket_endcopyout,
+	socket_nbytes_pending
 };
 
 const PQcommMethods *PqCommMethods = &PqCommSocketMethods;
@@ -1502,6 +1504,9 @@ socket_flush_if_writable(void)
 
 /* --------------------------------
  *	socket_is_send_pending	- is there any pending data in the output buffer?
+ *
+ * This doesn't report data pending in operating system socket buffers, only in
+ * the in-process libpq send buffer.
  * --------------------------------
  */
 static bool
@@ -1510,6 +1515,34 @@ socket_is_send_pending(void)
 	return (PqSendStart < PqSendPointer);
 }
 
+/* --------------------------------
+ *		pq_pending_size - report internal libpq buffer content sizes
+ *
+ * Reports size of data waiting to be sent to the OS in the internal libpq send
+ * buffer, if any. Does not include operating system socket buffering. Also
+ * reports libpq receive buffer content size, i.e. data read from OS socket but
+ * not yet consumed by application.
+ *
+ * Use socket_is_send_pending() if you just need to determine if there's data
+ * waiting to send.
+ *
+ * rxsz or txsz may be null if that data isn't required.
+ *
+ * Never fails on TCP sockets. The int return code is so that the wrapper
+ * function pq_nbytes_pending can return STATUS_ERROR for other socket types
+ * that don't support this.
+ */
+static int
+socket_nbytes_pending(size_t * rxsz, size_t * txsz)
+{
+	if (txsz)
+		*txsz = (PqSendStart < PqSendPointer) ? (PqSendPointer - PqSendStart) : 0;
+	if (rxsz)
+		*rxsz = PqRecvLength - PqRecvPointer;
+
+	return STATUS_OK;
+}
+
 /* --------------------------------
  * Message-level I/O routines begin here.
  *
diff --git a/src/backend/libpq/pqmq.c b/src/backend/libpq/pqmq.c
index f51d935daf..e4d6b7c779 100644
--- a/src/backend/libpq/pqmq.c
+++ b/src/backend/libpq/pqmq.c
@@ -44,7 +44,8 @@ static const PQcommMethods PqCommMqMethods = {
 	mq_putmessage,
 	mq_putmessage_noblock,
 	mq_startcopyout,
-	mq_endcopyout
+	mq_endcopyout,
+	NULL /* TODO nbytes_pending */
 };
 
 /*
diff --git a/src/include/libpq/libpq.h b/src/include/libpq/libpq.h
index b1152475ac..350febbfac 100644
--- a/src/include/libpq/libpq.h
+++ b/src/include/libpq/libpq.h
@@ -31,6 +31,7 @@ typedef struct
 	void		(*putmessage_noblock) (char msgtype, const char *s, size_t len);
 	void		(*startcopyout) (void);
 	void		(*endcopyout) (bool errorAbort);
+	int			(*nbytes_pending) (size_t *rxsz, size_t *txsz);
 } PQcommMethods;
 
 extern const PGDLLIMPORT PQcommMethods *PqCommMethods;
@@ -45,6 +46,32 @@ extern const PGDLLIMPORT PQcommMethods *PqCommMethods;
 	(PqCommMethods->putmessage_noblock(msgtype, s, len))
 #define pq_startcopyout() (PqCommMethods->startcopyout())
 #define pq_endcopyout(errorAbort) (PqCommMethods->endcopyout(errorAbort))
+#define pq_nbytes_pending(rxsz, txsz) ( PqCommMethods->nbytes_pending ? PqCommMethods->nbytes_pending((rxsz), (txsz)) : STATUS_ERROR )
+
+/*
+ * Socket statistics requests.
+ *
+ * TODO: Probably needs to become a VLA where we request the fields we want by
+ * supplying them tagged, and the callee fills in the values. That's very DSA
+ * friendly too.
+ */
+#define PQ_SOCKSTATS_TX_BUFSZ (1<<0)
+#define PQ_SOCKSTATS_RX_BUFSZ (1<<1)
+#define PQ_SOCKSTATS_TX_BUFCONTENTSZ (1<<2)
+#define PQ_SOCKSTATS_RX_BUFCONTENTSZ (1<<3)
+#define PQ_SOCKSTATS_TX_WINDOWSZ (1<<4)
+#define PQ_SOCKSTATS_RX_WINDOWSZ (1<<5)
+
+typedef struct PQsocketStats
+{
+	int32 set_fields;
+	size_t sock_tx_bufsz;				/* SO_SNDBUF */
+	size_t sock_tx_bufcontentsz;		/* SIOCOUTQ */
+	size_t sock_rx_bufsz;				/* SO_RCVBUF */
+	size_t sock_rx_bufcontentsz; 		/* SIOCINQ */
+	size_t sock_tx_windowsz;			/* scaled struct tcp_info.tcpi_snd_cwnd */
+	size_t sock_rx_windowsz;			/* struct tcp_info.tcpi_snd_wnd */
+} PQsocketStats;
 
 /*
  * External functions.
-- 
2.26.2

